#include "assembler.hpp"
#include "heatProblem.hpp"
#include <iostream>

HeatAssembler::HeatAssembler(Grid* g, HeatProblem* pb) {

  _grid = g;
  _pb   = pb;
  _rhs.resize(_grid->getNbPoints(0)*_grid->getNbPoints(1));
}

void HeatAssembler::assembleMatrix(double dt) {

  int Nx = _grid->getNbPoints(0);
  int Ny = _grid->getNbPoints(1);

  // We could use the () operator of the matrix, but it is really inefficient with large matrices.
  // Here we know the structure of the matrix, and what is exactly each coefficient,
  // so we will build the arrays of indices and coeffs then use the matrix constructor with vectors 
  std::vector<uint> rows,cols;
  std::vector<double> vals;
  uint row,col;

  double k  = _pb->getDiffusionCoeff();
  double dx = _grid->getSpacing(0);
  double dy = _grid->getSpacing(1);

  double invdxx = k*dt/(dx*dx);
  double invdyy = k*dt/(dy*dy);

  // i=0, j=0, line 0
  rows.push_back(0);
  cols.push_back(0);
  vals.push_back(1. + invdxx + invdyy);
  rows.push_back(0);
  cols.push_back(1);
  vals.push_back(-invdxx);
  rows.push_back(0);
  cols.push_back(Nx);
  vals.push_back(-invdyy);

  // 0<i<Nx-1, j=0
  for (int i=1;i<Nx-1;i++) {
    rows.push_back(i);
    cols.push_back(i-1);
    vals.push_back(-invdxx);
    rows.push_back(i);
    cols.push_back(i);
    vals.push_back(1. + 2*invdxx + invdyy);
    rows.push_back(i);
    cols.push_back(i+1);
    vals.push_back(-invdxx);
    rows.push_back(i);
    cols.push_back(i+Nx);
    vals.push_back(-invdyy);
  }

  // i=Nx-1,j=0
  rows.push_back(Nx-1);
  cols.push_back(Nx-1);
  vals.push_back(1. + invdxx + invdyy);
  rows.push_back(Nx-1);
  cols.push_back(Nx-2);
  vals.push_back(-invdxx);
  rows.push_back(Nx-1);
  cols.push_back(2*Nx-1);
  vals.push_back(-invdyy);

  for (int j=1;j<Ny-1;j++) {

    // i=0, 0<j<Ny-1
    int i=0;
    row = _grid->index1D(i,j);
    rows.push_back(row);
    cols.push_back(row-Nx);
    vals.push_back(-invdyy);
    rows.push_back(row);
    cols.push_back(row);
    vals.push_back(1. + invdxx + 2*invdyy);
    rows.push_back(row);
    cols.push_back(row+1);
    vals.push_back(-invdxx);
    rows.push_back(row);
    cols.push_back(row+Nx);
    vals.push_back(-invdyy);

    // 0<i<Nx-1, 0<j<Ny-1
    for (i=1;i<Nx-1;i++) {
      row = _grid->index1D(i,j);
      rows.push_back(row);
      cols.push_back(row-Nx);
      vals.push_back(-invdyy);
      rows.push_back(row);
      cols.push_back(row-1);
      vals.push_back(-invdxx);
      rows.push_back(row);
      cols.push_back(row);
      vals.push_back(1. + 2*invdxx + 2*invdyy);
      rows.push_back(row);
      cols.push_back(row+1);
      vals.push_back(-invdxx);
      rows.push_back(row);
      cols.push_back(row+Nx);
      vals.push_back(-invdyy);
    }

    // i=Nx-1, 0<j<Ny-1
    i = Nx-1;
    row = _grid->index1D(i,j);
    rows.push_back(row);
    cols.push_back(row-Nx);
    vals.push_back(-invdyy);
    rows.push_back(row);
    cols.push_back(row-1);
    vals.push_back(-invdxx);
    rows.push_back(row);
    cols.push_back(row);
    vals.push_back(1. + invdxx + 2*invdyy);
    rows.push_back(row);
    cols.push_back(row+Nx);
    vals.push_back(-invdyy);

  }

  // i=0, j=Ny-1
  row = _grid->index1D(0,Ny-1);
  rows.push_back(row);
  cols.push_back(row-Nx);
  vals.push_back(-invdyy);
  rows.push_back(row);
  cols.push_back(row);
  vals.push_back(1. + invdxx + invdyy);
  rows.push_back(row);
  cols.push_back(row+1);
  vals.push_back(-invdxx);

  // 0<i<Nx-1, j=Ny-1
  for (int i=1;i<Nx-1;i++) {
    row = _grid->index1D(i,Ny-1);
    rows.push_back(row);
    cols.push_back(row-Nx);
    vals.push_back(-invdyy);
    rows.push_back(row);
    cols.push_back(row-1);
    vals.push_back(-invdxx);
    rows.push_back(row);
    cols.push_back(row);
    vals.push_back(1. + 2*invdxx + invdyy);
    rows.push_back(row);
    cols.push_back(row+1);
    vals.push_back(-invdxx);
  }

  // i=Nx-1,j=Ny-1
  row = Nx*Ny-1;
  rows.push_back(row);
  cols.push_back(row-Nx);
  vals.push_back(-invdyy);
  rows.push_back(row);
  cols.push_back(row-1);
  vals.push_back(-invdxx);
  rows.push_back(row);
  cols.push_back(row);
  vals.push_back(1 + invdxx + invdyy);

  _A = SparseMatrix(rows,cols,vals);  

}

void HeatAssembler::assembleRHS(rvec& un, double t, double dt) {

  int Nx = _grid->getNbPoints(0);
  int Ny = _grid->getNbPoints(1);
  for (int i=0;i<Nx;i++)
    for (int j=0;j<Ny;j++) {
      int row = _grid->index1D(i,j);
      double x,y;
      _grid->coords(i,j,x,y);
      _rhs[row] = un[row] + dt*_pb->sourceTerm(x,y,t);
  }

}
