#include "linearSolver.hpp"
#include <iostream>
#include <unistd.h>
#include <cmath>

LinearSolver::LinearSolver(Parameters* p, Grid* g) {
                            
  _params = p;
  _grid   = g;                                                      

  std::string stype = p->getString("linear solver","CG");

  if (stype != "CG" and stype != "LU") {
    std::cerr << "LinearSolver: unknown solver type " << stype << std::endl;
    std::cerr << "Valid types are CG and LU" << std::endl;
    abort();
  } 

  if (stype == "CG") _type = solverType::CG;
  else _type = solverType::DIRECT;

    
}

void LinearSolver::solve(SparseMatrix& A, rvec& x, rvec& b) {

  switch (_type) {
    case solverType::CG:
      solveCG(A,b,x);
      break;
    case solverType::DIRECT:
      std::cerr<<"Not implemented" << std::endl;
      abort();
      break;
    default:
      std::cerr << "LinearSolver::solver: wrong solver type stored in class instance.";
      abort();
  }

}

void LinearSolver::solveCG(SparseMatrix& A, rvec& x, rvec& b) {

  // Check sizes
  if (A.size() != b.size() or x.size() != A.size()) {
std::cout << A.size() << " " << b.size() << " " << x.size() << std::endl;
    std::cerr << "solveCG: incompatible sizes of vectors and matrix " << std::endl;
    abort();
  }

  double relTol  = _params->getReal("rtol",1.e-3);
  double absTol  = _params->getReal("atol",1.e-6);
  double normR   = 1.e3,normRPrev;
  double diff    = 1.e3;
  int    iterMax = 1000;
  int    iter    = 0;

  rvec residual (A.size());
  rvec residual2(A.size());
  rvec tmp(A.size());
  rvec direction(A.size());

  // Init: we start from x0 = 0, so first residual is just b
  // First direction is b as well
  for (uint i=0u;i<A.size();i++) {
    x        [i] = 0.0;
    residual [i] = b[i];
    direction[i] = residual[i];
  }

  do {

    normRPrev = normR;
    tmp = A*direction;
    double distance = dotProd(residual,residual)/dotProd(direction,tmp);
    normR = 0.e0;
    for (uint i=0u;i<A.size();i++) {
      x[i]         += distance*direction[i];
      residual2[i]  = residual[i] - distance*tmp[i];
      normR        += pow(residual2[i],2);
    }
    normR = sqrt(normR);
    double distance2 = dotProd(residual2,residual2)/dotProd(residual,residual);
    for (uint i=0u;i<A.size();i++) {
      direction[i] = residual2[i] + distance2*direction[i];
      residual [i] = residual2[i];
    } 

    diff = fabs(normRPrev-normR)/normRPrev;
    iter ++;
  } while (diff > relTol and normR > absTol and iter < iterMax); 

  std::cout << "CG finished after " << iter << " iterations." << std::endl;


}
